%******************************************************************************;
%* SAS macro Chk4UV   vs. 0.4.0 , 28 Jun 2001, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%*                                                                            *;
   %MACRO   CHK4UV     /* Check for Unique Values in multiple partial records */
/* -------------------------------------------------------------------------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  SourceDs=_LAST_     /* Source data set specification                     R */
,  IdList  =           /* Variable List of which multiple Identical value   R */
                       /*     combinations determine a single new record      */
                       /*     (often variables like a Case Number)            */
,  IdFormat=UNFORMATTED/* identify cases by unformatted values of identifi- O */
                       /*     cation variables, alternative argument value:   */
                       /*   FORMATTED (or explicitely empty) : identify       */
                       /*     cases by eventually formatted values of         */
                       /*     identification variables                        */
,  MRList  =           /* Multiple Records variable name List to keep and   O */
                       /*     optionally check while keeping unique values,   */
                       /*     No value or '_ALL_' involves all variables      */
,  Check   =           /* Name of global macro variable with check result=  R */
                       /*    OK:   successful compression to unique values    */
                       /*    NO: unsuccessful compression to unique values    */
                       /*   no global macro variable name: no check at all    */
,  Index   =           /* Indexing measures variable name: numeric integer  O */
                       /*     and unique within IdList, default automatic ~)  */
,  Lowest  =           /*  Lowest (integer) value of index variable         O */
,  Highest =           /* Highest (integer) value of index variable         O */
,  DupAsOne=N          /* consider duplicate MRlist values as the same one  O */
                       /* while checking for multiple values within IdList    */
,  KeepDups=N          /* keep duplicate records (equal index values): Y/N  O */
,  ForMyUse=           /* argument for the author's use only, do not use it ! */
)/ DES     =             'Check for Unique,exclusive Values in MPR'/* STORE */ ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : CHK4UV.SAS / CHK4UV.040 ( / CHK4UV__.SAS )                    *;
%* Directory  : .................                                             *;
%*                                                                            *;
%* Purpose    : Checking multiple partial record structure, Unique Values     *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 28 Jun 2001                                                   *;
%* Version    : 0.40                                                          *;
%*                                                                            *;
%* SAS version: 6.12 under Windows 95                                         *;
%*                                                                            *;
%* Validation : -                                                             *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Notes:                                                                     *;
%* ======                                                                     *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of Chk4UV:                                                         *;
%* ==================                                                         *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.4      28 Jun 2001  checking part splitted from MPR2SR.SAS, 1st release  *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* remove comment chars for testing ;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN
  %DO;
    %*ValueLst (ForMyUse=Compile);
    %GOTO Finish;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* LOCAL macro variables declarations;
%*___________________________________;

 %LOCAL    /* Define internal macro variables explicitely local */
  MacName  /* This macros external name: Chk4UV */
  I        /* element index */
  Id       /* IdList element */
  ByList   /* all IdList elements eventually preceeded by "GROUPFORMAT" */
  IdFrqLst /* IdList with elements separated by singele *'s indicating tables */
  LastId   /* last Id from IdList */
  MultRec  /* MRlist element */
  _Sorted1 /* stand-in name for sorted dataset 1 XX_Sort1 */
  _Sorted2 /* stand-in name for sorted dataset 2 XX_Sort2 */
  _Extreme /* stand-in name for Extremes dataset XX_Extr_ */
  Contents /* stand-in name for Contents dataset XX_Cont_ */
  Scratch  /* stand-in name for Scratch datasets XX_Scrat */
  ValList  /* temporary variable value list storage */
  VarValue /* temporary variable value storage, element of ValList */
  NrNonMis /* temporary variable NrNonMis (number of non-missings) storage */
  ErrCount /* number of erroneously specified arguments */
  Err      /* text 'E R R O R' as consecutive characters */
  Warn     /* text 'W A R N I N G' as consecutive characters */
  ;

%* Stand-in variable and dataset names;
  %LET MacName  = Chk4UV  ;
  %LET _Sorted1 = XX_Sort1;
  %LET _Sorted2 = XX_Sort2;
  %LET _Extreme = XX_Extr_;
  %LET Contents = XX_Cont_;
  %LET Scratch  = XX_Scrat;
  %LET ValList  = XX_VaLst;
  %LET VarValue = XX_Value;
  %LET NrNonMis = XX_VarNr;
  %LET Err      = %SYSFUNC(COMPRESS(E R R O R, %STR( )));
  %LET Warn     = %SYSFUNC(COMPRESS(W A R N I N G, %STR( )));

%* Macro header in LOG file;
  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs. 0.4, by Jim Groeneveld, 28 Jun 2001   |;
  %PUT |    Check for unique Multiple Partial Records structure    |;
  %PUT |___________________________________________________________|;
  %PUT;

%* Feedback of specified arguments;
  %PUT &MacName: SourceDs = &SourceDs;
  %PUT &MacName: IdList   = &IdList  ;
  %PUT &MacName: IdFormat = &IdFormat;
  %PUT &MacName: MRlist   = &MRlist  ;
  %PUT &MacName: Check    = &Check   ;
  %PUT &MacName: Index    = &Index   ;
  %PUT &MacName: Lowest   = &Lowest  ;
  %PUT &MacName: Highest  = &Highest ;
  %PUT &MacName: DupAsOne = &DupAsOne;
  %PUT &MacName: KeepDups = &KeepDups;
  %*PUT &MacName: ForMyUse = &ForMyUse;

%* Convert lower case letters to upper case and keep first argument characters;
  %IF ( &DupAsOne NE ) %THEN %LET DupAsOne = %UPCASE(%SUBSTR(&DupAsOne,1,1));
  %IF ( &KeepDups NE ) %THEN %LET KeepDups = %UPCASE(%SUBSTR(&KeepDups,1,1));

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Check arguments for legal contents and discrepancies, calling e_r_r_o_r_s    ;
*______________________________________________________________________________;

  %LET ErrCount = 0;
  %IF ( &Lowest NE AND &Highest NE ) %THEN
  %DO;
    %IF ( &Lowest > &Highest ) %THEN
    %DO;
      %PUT *** &MacName *** &ERR: Lowest (&Lowest) larger than Highest (&Highest),;
      %PUT                       macro &MacName will abort;
      %LET ErrCount = %EVAL ( &ErrCount + 1);
    %END;
  %END;
  %IF ( &SourceDs EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: SourceDs argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&SourceDs)) EQ 0 ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Dataset &SourceDs does not exist,;
    %PUT                       macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %ELSE
  %DO; %* Check for non-existing variables within dataset;
   %IF ( %UPCASE(&MRlist) NE %STR(_ALL_) AND &MRlist NE ) %THEN
    %ChkVar (Data=&SourceDs, VarList=&IdList &MRlist &Index, CalledBy=&MacName,
             Result = ErrCount); %* adding up non-existent variables ;
   %ELSE
    %ChkVar (Data=&SourceDs, VarList=&IdList &Index, CalledBy=&MacName,
             Result = ErrCount); %* adding up non-existent variables ;
  %END;
  %IF ( &IdList EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: IdList argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
  %IF ( &Check EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Check argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
/*=========
  %IF ( &MRlist EQ ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: MRlist argument not specified, macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1);
  %END;
=========*/
  %IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Determine last Id variable from IdList and build BY list with "GROUPFORMAT" ;
%*_____________________________________________________________________________;

  %IF ( %UPCASE(&IdFormat) NE UNFORMATTED ) %THEN %LET ByList = ;
  %LET I = 1;
  %LET Id = %SCAN(&IdList,&I,%STR( ));
  %DO %WHILE (&Id NE);
    %LET LastId = &Id;
    %IF ( %UPCASE(&IdFormat) NE UNFORMATTED ) %THEN
      %LET ByList = &ByList GROUPFORMAT &Id;
    %LET I = %EVAL( &I + 1 );
    %LET Id = %SCAN(&IdList,&I,%STR( ));
  %END; %*=== While;

  %IF ( %UPCASE(&IdFormat) EQ UNFORMATTED ) %THEN %LET ByList = &IdList;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Build MRlist from keyword _ALL_, exclude IdList and Index variables ;
%*_____________________________________________________________________;

%IF (%UPCASE(&MRlist) EQ %STR(_ALL_) OR %UPCASE(&MRlist) EQ /*empty*/) %THEN
%DO;
  %LET MRlist = ; %* empty;
  PROC CONTENTS DATA=&SourceDs (DROP=&IdList &Index)
                 OUT=&Contents (KEEP=Name) NOPRINT;
  RUN;

  DATA _NULL_;
    SET &Contents;
    CALL EXECUTE ( '%LET MRlist = &MRlist ' || Name || ';' );
  RUN;
%PUT &MacName: Generated MRlist=&MRlist;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic determination of attributes and check on numeric index variable ;
%*___________________________________________________________________________;

PROC CONTENTS DATA=&SourceDs (KEEP=&MRList &Index) OUT=&Contents NOPRINT;
RUN;

DATA _NULL_;
  SET &Contents (KEEP=Name Label Informat Format Type Length);

 %LET I = 1;
 %LET MultRec = %SCAN(&MRList,&I,%STR( ));
 %DO %WHILE (&MultRec NE);

   %IF ( &I GT 1 ) %THEN ELSE;
    IF ( Name = UPCASE("&MultRec") ) THEN
    DO;
%* Make attributes as macro variables;
     %LOCAL MR&I.TI; %* Define locals;
      CALL SYMPUT ("MR&I.T", TRIM(LEFT(PUT(Type,1.0))));
    END;

    %LET I = %EVAL( &I + 1 );
    %LET MultRec = %SCAN(&MRList,&I,%STR( ));
  %END; %*=== While;

  %IF ( &Index NE ) %THEN
  %DO;
    IF ( Name = UPCASE("&Index") ) THEN
    DO;
      IF ( Type NE 1 ) THEN
      DO;
        PUT "*** &MacName *** &ERR: Index variable &Index is not numeric,";
        PUT "                      macro &MacName will abort";
        CALL SYMPUT ("ErrCount", "1");
      END;
    END;
  %END;
RUN;                  %* @@@@@@@@@ Why not character index variable? @@@@@@@@@;

PROC DATASETS NOLIST;
  DELETE &Contents;                        %* Delete scratch dataset &Contents;
RUN;

%IF ( &ErrCount GT 0 ) %THEN %GOTO Interupt;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic generation of index variable (if none specified) ;
%*____________________________________________________________;

%IF ( &Index EQ ) %THEN
%DO;
  %LET Index = XX_Index; %* XX_Index assumed unique temporary name;
  PROC SORT DATA=&SourceDs (KEEP=&IdList &MRlist)
    OUT=&_Sorted1; BY &IdList;   %* Sort input file;
   %IF ( %UPCASE(&IdFormat) EQ UNFORMATTED ) %THEN %STR(FORMAT &IdList;) ;
  RUN;
  DATA &_Sorted1;
    SET &_Sorted1; BY &ByList;
    IF ( First.&LastId ) THEN &Index = %IF ( &Lowest NE ) %THEN &Lowest; %ELSE 1;;
    ELSE &Index + 1;
  RUN;
  %LET SourceDs = &_Sorted1; %* let SourceDs point to the resulting dataset;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Automatic determination of minimum and maximum Index values ;
%*_____________________________________________________________;

%IF ( &Lowest EQ OR &Highest EQ ) %THEN
%DO;
  PROC UNIVARIATE DATA=&SourceDs (KEEP=&Index) NOPRINT;
    VAR &Index;
    OUTPUT OUT = &_Extreme
           MIN = Minimum
           MAX = Maximum ;
  RUN;

  DATA _NULL_;
    SET &_Extreme;
    %IF ( &Lowest  EQ ) %THEN CALL SYMPUT ("Lowest" , TRIM(LEFT(PUT(Minimum,9.0))));;
    %IF ( &Highest EQ ) %THEN CALL SYMPUT ("Highest", TRIM(LEFT(PUT(Maximum,9.0))));;
  RUN;
%END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Convert eventual real Lowest and Highest values to integers ;
%*_____________________________________________________________;

 %IF (&Lowest  NE .) %THEN  %LET Lowest   = %SYSEVALF(&Lowest ,CEIL);
 %IF (&Highest NE .) %THEN  %LET Highest  = %SYSEVALF(&Highest,CEIL);

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check lowest and highest values ;
%*_________________________________;

%IF ( &Lowest > &Highest ) %THEN
%DO;
  %PUT *** &MacName *** &ERR: Lowest (&Lowest) larger than Highest (&Highest),;
  %PUT                      macro &MacName will abort;
  %GOTO Aborting;
%END;
%ELSE %PUT Lowest index value=&Lowest and Highest index value=&Highest;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro actually starts here ;
%*____________________________;

%* Presort dataset according to &IdList and &Index, argument dependent;
PROC SORT DATA=&SourceDs (KEEP=&IdList &MRlist &Index)
  %IF ( &KeepDups=N ) %THEN
  %DO;
    NODUPKEY
  %END;
    OUT=&_Sorted2; BY &IdList &Index;  %* Sort input file;
  %IF ( %UPCASE(&IdFormat) EQ UNFORMATTED ) %THEN %STR(FORMAT &IdList;) ;
  FORMAT &Index;             %* Dissociate eventual format from index variable;
RUN;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Check for a single non-missing value at maximum across multiple records ;
%*_________________________________________________________________________;

%* Make IdList with multiplication (*) characters in between;
  %LET IdFrqLst = ;
  %LET I = 1;
  %LET Id = %SCAN(&IdList,&I,%STR( ));
  %DO %WHILE (&Id NE);
    %LET IdFrqLst = &IdFrqLst &Id *;
    %LET I = %EVAL( &I + 1 );
    %LET Id = %SCAN(&IdList,&I,%STR( ));
  %END; %*=== While;

  %LET &&Check = OK;

  %LET I = 1;
  %LET MultRec = %SCAN(&MRList,&I,%STR( ));
  %DO %WHILE (&&&Check EQ OK AND &MultRec NE);

    PROC FREQ DATA=&_Sorted2 (KEEP=&IdList &MultRec &Index
      WHERE=(&Index LE .z OR (&Index GE &lowest AND &Index LE &highest)) );
      TABLES &IdFrqLst &MultRec / NOPRINT NOPERCENT
        OUT=&Scratch (KEEP=&IdList &MultRec Count);
      RUN;

    DATA _NULL_;
      SET &Scratch; BY &IdList; * Necessary for use with First.BY and Last.BY;
      IF (First.&LastId) THEN &NrNonMis = 0;
     %IF &&MR&I.T EQ 1 %THEN /* numeric variable */
     %DO;
      IF &MultRec NE . THEN  /* Don't count missing values */
           /* This excludes all different missing values .a-.z */
     %END;
     %ELSE /* character variable */
     %DO;
      IF &MultRec NE ' ' THEN  /* Don't count missing values */
     %END;
      DO;
        &NrNonMis + 1;
        IF ( %IF (&DupAsOne EQ N) %THEN (&NrNonMis EQ 1 AND Count GT 1) OR;
        &NrNonMis EQ 2) THEN
        DO;
          CALL EXECUTE ( '%LET &&Check = NO;' );
          PUT "&Check=NO, no unique values found for each &IdList combination";
          STOP;
        END;
      END;
    RUN;

    PROC DATASETS NOLIST;
      DELETE &Scratch;   %* Delete scratch datasets;
    RUN;

    %LET I = %EVAL( &I + 1 );
    %LET MultRec = %SCAN(&MRList,&I,%STR( ));
  %END; %*=== While;

%Aborting:

PROC DATASETS NOLIST;
  DELETE &_Sorted1 &_Sorted2 &_Extreme;   %* Delete scratch datasets;
RUN;

%Interupt:

 %IF ( &ErrCount GT 0 ) %THEN
 %DO;
   DATA _NULL_;
     ABORT /*ABEND*/;
   RUN;
 %END;

%Finish:

%***********;
%MEND Chk4UV; * --------------------end-of-macro-Chk4UV------------------------;
************;

